const _ = require('lodash');
const stmTypes = require('../core/statement');
const kinds = require('./kinds');

let keepComment = false;
let autoMock = false;
let defaultKind = 'ajax';

function transform(ast, options) {
  autoMock = options.mock || options.autoMock;
  keepComment = options.keepComment;
  defaultKind = kinds.defaultKind;
  const usedKinds = {};
  let imports = [];
  const variables = [];
  const body = [];
  let defaultExport = '';
  const typeNInterface = [];

  ast.statements.forEach((statement) => {
    if (statement.ignore) {
      body.push(statement.text);
    } else if (statement instanceof stmTypes.Clazz) {
      const mockClzName = _.get(statement.name.match(/(\w+)Mock$/), '1');
      if (mockClzName && !autoMock && !!_.find(ast.statements, (st) => st.name === mockClzName)) {
        return;
      }
      if (statement.modifiers.includes('declare')) {
        const hasMockClz = !!_.find(ast.statements, (st) => st.name === `${statement.name}Mock`);
        body.push(makeImplementClz(statement, hasMockClz, usedKinds));
      } else {
        body.push(statement.text);
      }
    } else if (statement.isFunc) {
      const typeDef = _.find(ast.statements, (item) => {
        if (item === statement) return false;
        return item.name === statement.name || item.name + 'Validator' === statement.name;
      });
      if (typeDef) {
        typeDef.hasDataCheck = typeDef.name === statement.name ? 1 : 2;
        variables.push(typeDef.hasDataCheck === 2 ? statement.text : renameTypeCheckName(statement));
      } else {
        body.push(statement.text);
      }
    } else if (statement instanceof stmTypes.Import) {
      imports.push(statement.text);
    } else if (statement instanceof stmTypes.Interface || statement instanceof stmTypes.TypeAlias || statement instanceof stmTypes.IntersectionType || statement instanceof stmTypes.TypeReference) {
      body.push(statement.text);
      typeNInterface.push(statement);
    } else if (statement instanceof stmTypes.Export) {
      defaultExport = statement.text;
    } else if (!(statement instanceof stmTypes.EndOfFile)) {
      body.push(statement.text);
    }
  });

  const usedSolutionKeys = _.keys(usedKinds);
  const importDeclare = kinds.genImportDeclare(usedSolutionKeys, ast.fileName, autoMock);
  imports = [].concat(importDeclare.imports, imports);

  if (usedSolutionKeys.length) {
    const mws = [].concat(importDeclare.mws);
    variables.push(`const _requestHandles: any = {${[0].concat(mws).reduce((prev, cur) => `${prev ? prev+',':''} ${cur}: ${cur}_kind`)} };`);
    variables.push(`const _toStepError = (err: any, step: string, data: any) => {err.step = step; err.data = data; return err}`)
  }
  
  const combined = [].concat(
    '// @ts-nocheck',
    imports,
    '\n',
    variables,
    '\n\n',
    body,
    '\n',
    makeTypeDeclare(typeNInterface),
    '\n',
    defaultExport,
  );
  return combined.join('\n').replace(/\n{2,}/g, '\n\n').replace(/^\n+/, '');
}

function makeImplementClz(clz, hasMockClz, usedKinds) {
  const hasCustomMock = autoMock && hasMockClz;
  return `${clz.isExport ? 'export ' : ''}class ${clz.name} {
  _config = ${JSON.stringify(clz.config)}
  ${!hasCustomMock ? '' : `customMock = new ${clz.name}Mock();`}
  ${clz.decorators ? `constructor(){
    ${clz.heritage ? 'super()\n' : ''}${makeDecorator(clz.decorators, true)}
  }` : ''}
  ${clz.methods.map((method) => {
    const config = method.config;
    const kind = config.kind || clz.config.kind || defaultKind;
    if (!autoMock) {
      delete config.mock;
    }
    const argsStr = method.parameters.map((p) => p.name).join(', ');
    const mockPipe = autoMock ? kinds.kinds._mock || [] : [];
    // 是否存在预定义的管道类型,如果不存在则当做是运行时注入处理
    let pipe = kinds.kinds[kind];
    if (!pipe || !pipe.length) {
      return `${method.name}(${method.parameters.map((p) => p.text).join(', ')}): ${method.returnType.text} {
  throw new Error('request kind ${kind} should be register');
}`;
    }
    usedKinds[kind] = true;
    return `${method.name}(${method.parameters.map((p) => p.text).join(', ')}): ${method.returnType.text} {
    const _param: any = {${argsStr}};
    const _paramSchema: any = ${type2Code(method.parameters, 0)};
    const _returnSchema: any = ${type2Code(_.get(method.returnType, 'typeArguments[0]', _.omit(method.returnType, 'config')), 0)};
    const _config: any = Object.assign({}, this._config, ${JSON.stringify(config)});
    const _callee: any = '${clz.name}.${method.name}';
    const _mwArgs: any = { param: _param, config: _config, paramSchema: _paramSchema, returnSchema: _returnSchema, callee: _callee, instance: this };
    ${makeDecorator(method.decorators)}
    // @ts-ignore
    ${!mockPipe.length ? '' : `if (window.__mock__ || _config.mock) {
      ${makePromisePipe(mockPipe, 3)}
    }`}
    ${makePromisePipe(pipe, 1)}
  }`;
  }).join('\n  ')}
}`;
}

function makePromisePipe(pipe, indentOffset) {
  const wrap = (i) => `${indent(i + 1, indentOffset)}Promise.resolve(_requestHandles.${pipe[i].name}(_mwArgs, ${i > 0 ? `r${i - 1}` : '_param'})).then((r${i}: any) => {
    ${i < pipe.length - 1 ? wrap(i + 1) : `${indent(i + 2, indentOffset)}resolve(r${i});`}
    ${indent(i + 1, indentOffset)}}).catch((e: any) => {
      ${indent(i + 1, indentOffset)}reject(_toStepError(e, '${pipe[i].name}', ${i > 0 ? `r${i - 1}` : '_param'}));
    ${indent(i + 1, indentOffset)}})`;
  return `return new Promise((resolve, reject) => {
    ${wrap(0)}
    ${indent(0, indentOffset)}})`;
}

function makeTypeDeclare(typeDefineds) {
  if (!typeDefineds.length) return;
  const declareSentence = 'const '+ _.map(typeDefineds, d => `_${d.name} = {}`).join(', ');
  const assignValue = _.map(typeDefineds, (def) => {
    const prefix = def.isExport ? 'export ' : '';
    return `${prefix}const ${def.name}:any = Object.assign(_${def.name}, ${type2Code(def, 2, def.hasDataCheck, _.map(typeDefineds, d => d.name))});`;
  }).join('\n');
  return declareSentence + '\n' + assignValue;
}

function renameTypeCheckName(func) {
  return func.text.replace(func.name, func.name + 'Validator').trim();
}

function makeModule(mod) {
  if (!mod) return '';
  return mod.map((subAst) => {
    const moduleDef = {};
    subAst.statements.forEach((st) => {
      if (st.statement.type) {
        moduleDef[st.statement.name] = `#${st.statement.type}#`;
      } else {
        moduleDef[st.statement.name] = st.statement;
      }
    });
    return `const ${subAst.name} = ${type2Code(moduleDef)}`;
  }).join('\n');
}

function type2Code(typedef, tab, hasCustomValidator, currentFileTypeDeclare) {
  if (!currentFileTypeDeclare) currentFileTypeDeclare = [];
  const simplifyStruct = (def) => {
    let rlt = { type: 'Object' };
    [
      'isOptional',
      'isArray',
      'typeArguments',
      keepComment ? 'comment' : '',
      'config',
    ].forEach((k) => {
      if (!k || !def[k]) return;
      rlt[k] = def[k];
    });
    if (def.comment) {
      const strValidator = def.comment.match(/`(.+)`/);
      if (strValidator && strValidator[1]) {
        rlt['validator'] = strValidator[1].replace(/\\/g, '\\\\');
      }
    }
    if (def instanceof stmTypes.Literal && def.literal) {
      // 字面量
      rlt['type'] = 'Literal';
      rlt['properties'] = def.type;
    } else if (def instanceof stmTypes.IntersectionType) {
      // A & B 语法
      rlt['type'] = 'Merge';
      rlt.properties = def.map((t) => simplifyStruct(t));
      return rlt;
    } else if (def instanceof stmTypes.UnionType) {
      // 枚举列表
      rlt['type'] = 'Enum';
      rlt['properties'] = def.map((t, i) => simplifyStruct(t));
    } else if (def.type === 'Array') {
      // Array<> 泛型特殊处理
      rlt = simplifyStruct(def.typeArguments[0]);
      rlt['isArray'] = true;
    } else if (def.isRefer) {
      // 类型引用
      const referType = def.type.type || def.type;
      rlt['type'] = 'Refer';
      rlt['properties'] = currentFileTypeDeclare.indexOf(referType) !== -1 ? `#_${referType}#` : `#${referType}#`;
    } else if (def.type) {
      rlt['type'] = def.type;
    }
    if (def.length && !rlt.properties) {
      rlt.properties = {};
      def.forEach((t) => {
        rlt.properties[t.name] = simplifyStruct(t);
      });
    }
    if (def.heritage) {
      rlt.extend = def.heritage.map(d => `#${d.name}#`)
    }
    return rlt;
  };
  const printObj = json => {
    return JSON.stringify(json, undefined, tab || 0)
    .replace(/"#(.+?)#"/g, '$1')
    .replace(/\\\\/g, '\\')
  }
  const finalJSON = simplifyStruct(typedef, []);
  if (hasCustomValidator) {
    finalJSON['validator'] = `#${typedef.name}Validator#`;
  }
  return printObj(finalJSON);
}

const indent = (n, offset) => new Array(n * 2 + offset).join(' ');

const makeDecorator = (decorators, isClz) => {
  if (!decorators || !decorators.length) return '';
  return decorators.map(d => {
    return `${d.expression}(${d.args.concat(isClz ? 'this._config' : '_config').join(', ')})`
  }).join(';\n  ').replace(/(\r)?\n/g, '\n  ') + ';';
}

module.exports = transform;
