const ts = require('typescript');
const _ = require('lodash');
const { getKindStr, getLiteralValue, isExport, isArray, isSimpleType, setOrPush, getRealKind } = require('./utils');

const { SyntaxKind } = ts;

function _drawCommonProp(ctx, node) {
  ctx.isArray = isArray(node);
  ctx.isExport = isExport(node);
  ctx.text = node.getFullText().trim();
  ctx.ignore = /^\/\/ *@api-ignore/.test(ctx.text);
  ctx.type = getKindStr(node);
  ctx.isRefer = getRealKind(node) === SyntaxKind.TypeReference;
  ctx.isOptional = !!node.questionToken;
  if (node.name) {
    ctx.name = node.name.getText().trim();
  } else if (node.typeName) {
    ctx.name = node.typeName.text;
  } else {
    ctx.name = _.result(node, 'parent.name.getText');
  }
  const jsDoc = node.jsDoc;
  ctx.comment = _.result(node, 'jsDoc[0].getText');
  _.get(jsDoc, '0.tags', []).forEach((tag) => {
    if (!ctx.config) ctx.config = {};
    setOrPush(ctx.config, _.get(tag.name || tag.tagName, 'text'), tag.comment);
  });
}

function Literal(node) {
  _drawCommonProp(this, node);
  // 引用类型
  if (node.kind === SyntaxKind.TypeReference) {
    this.type = getKindStr(node);
    return;
  }
  // 字面量类型
  if (node.kind === SyntaxKind.LiteralType) {
    this.literal = true;
    this.type = getLiteralValue(node, node.literal.kind);
  }
  if (this.type === undefined) {
    this.type = getKindStr(node);
  }
  this.isSimple = this.literal || isSimpleType(node);
}

/** 或组合类型 */
function UnionType(node) {
  node = node.type || node;
  const kind = getRealKind(node);
  if (kind === SyntaxKind.FunctionType) {
    return new Func(node);
  }
  if (kind === SyntaxKind.TypeReference) {
    return new TypeReference(node);
  }
  if (kind !== SyntaxKind.UnionType) {
    return getKindStr(node);
  }
  if (!(this instanceof UnionType)) {
    return new UnionType(node);
  }
  _drawCommonProp(this, node);
  (node.type || node).types.forEach((t) => this.push(new Literal(t)));
}
UnionType.prototype = [];

/** 引用类型 */
function TypeReference(node) {
  if (getRealKind(node) !== SyntaxKind.TypeReference) return;
  if (!(this instanceof TypeReference)) {
    return new TypeReference(node);
  }
  _drawCommonProp(this, node);
  this.typeArguments = TypeArgument(node);
}

/** 泛型参数 */
function TypeArgument(node) {
  if (!node) return;
  const typeArguments = (node.type || node).typeArguments;
  if (!typeArguments) return;
  if (!(this instanceof TypeArgument)) {
    return new TypeArgument(node);
  }
  typeArguments.forEach((sub) => {
    this.push(visitNode(sub));
  });
}
TypeArgument.prototype = [];

function IntersectionType(node) {
  // type = xx & yy;
  (node.types || node.type.types).forEach(n => {
    this.push(visitNode(n));
  })
  _drawCommonProp(this, node);
}
IntersectionType.prototype = [];

/** type alias */
function TypeAlias(node) {
  if (!node) return;
  const kind = getRealKind(node);
  if (node.kind === SyntaxKind.PropertySignature) {
    const rtl = TypeAlias(node.type);
    rtl.isOptional = !!node.questionToken;
    rtl.name = node.name.getText();
    _drawCommonProp(rtl, node);
    return rtl;
  }
  if (kind === SyntaxKind.VoidKeyword) {
    return null;
  }
  if (kind === SyntaxKind.UnionType) {
    return UnionType(node)
  }
  if (!(this instanceof TypeAlias)) {
    return new TypeAlias(node);
  }
  _drawCommonProp(this, node);
  this.isArray = _.includes([SyntaxKind.TupleType, SyntaxKind.ArrayType], _.get(node.type, 'kind', node.kind));
  const members = _.get(node, 'type.members') || _.get(node, 'type.elementType.members') || node.members;
  if (members) {
    // eslint-disable-next-line no-restricted-syntax
    for (const child of members) {
      this.push(TypeAlias(child));
    }
    this.text = node.getText().trim();
    this.isOptional = !!node.questionToken;
  } else {
    this.isSimple = isSimpleType(node);
    this.type = UnionType(node);
    if (node.type && node.type.typeArguments) {
      this.typeArguments = TypeArgument(node);
    }
  }
}
TypeAlias.prototype = [];

/** interface */
function Interface(node) {
  _drawCommonProp(this, node);
  node.members.forEach((child) => {
    this.push(TypeAlias(child));
  });
  // const extendI = _.get(node, 'heritageClauses[0].types[0]');
  this.heritage = !node.heritageClauses ? null : node.heritageClauses[0].types.map(t => ({
    name: t.expression.getText(),
    start: t.getFullStart() - node.getFullStart(),
    end: t.getEnd() - node.getFullStart(),
    typeArguments: TypeArgument(t),
  }))
}
Interface.prototype = [];

/** 变量 */
function Variable(node) {
  const delaration = node.declarationList.declarations[0];
  this.name = delaration.name.text;
  this.text = node.getText();
  this.nameStart = 6;
  this.isExport = isExport(node);
  this.node = node;
  this.isFunc = [SyntaxKind.ArrowFunction, SyntaxKind.FunctionExpression, SyntaxKind.FunctionType].indexOf((delaration.initializer || delaration.type).kind) !== -1;
}

/** 函数 */
function Func(node) {
  _drawCommonProp(this, node);
  if (node.name) {
    this.nameStart = node.name.getStart() - node.getStart();
  }
  this.parameters = [];
  (node.parameters || []).forEach((arg) => {
    if (!arg.type) {
      this.parameters[arg.name.text] = {};
      return;
    }
    this.parameters.push(new MethodParameter(arg, this.config));
  });
  this.returnType = visitNode(node.type);
}

/** 模块 module/namespace */
function SubModule(node) {
  const sub = {};
  node.symbol.exports.forEach((symbol) => {
    sub[node.name.getText()] = visitNode(symbol.declarations[0]);
  });
  return sub;
}

/** method arguments */
function MethodParameter(node, paramDoc) {
  if (!paramDoc) paramDoc = {};
  _drawCommonProp(this, node);
  this.comment = paramDoc[node.name.text];
  delete paramDoc[node.name.text];
  this.isSimple = isSimpleType(node);
  this.isOptional = !!node.questionToken || undefined;
  this.type = UnionType(node);
}

function Decorator(decorators) {
  if (!decorators || !decorators.length) {
    return
  }
  if (!(this instanceof Decorator)) {
    return new Decorator(decorators);
  }
  decorators.forEach(d => {
    this.push({
      text: d.getText().substr(1),
      expression: (d.expression.expression || d.expression).getText(),
      args: !d.expression.arguments ? null : d.expression.arguments.map(a => a.getText()),
    });
  });
}
Decorator.prototype = [];

/** Class method */
function ClzMethod(node) {
  const paramDesc = {};
  const valueDeclaration = node.symbol.valueDeclaration;
  /** doc 注释文档 */
  this.doc = _.get(valueDeclaration.jsDoc, '0.comment');
  /** 形名 */
  this.name = node.name.getText();
  /** 注释中的`@xx`属性 */
  this.config = {};
  this.body = _.result(node, 'valueDeclaration.body.getFullText');
  this.decorators = Decorator(node.decorators);
  /** 返回类型 */
  this.returnType = visitNode(valueDeclaration.type);
  this.nameStart = valueDeclaration.name.getStart() - valueDeclaration.getFullStart();
  this.text = valueDeclaration.getText().trim();
  this.fulltext = valueDeclaration.getText().trim();
  _.get(valueDeclaration, 'jsDoc[0].tags', []).forEach((t) => {
    if (t.kind === SyntaxKind.JSDocParameterTag) {
      paramDesc[t.name.text] = t.comment;
      return;
    }
    setOrPush(this.config, t.tagName.text, t.comment);
  });
  /** 入参参数 */
  this.parameters = valueDeclaration.parameters.map((arg) => {
    return new MethodParameter(arg, paramDesc);
  });
}

/** Class 声明 */
function Clazz(node) {
  const symbol = node.symbol;
  const extendClz = _.get(node, 'heritageClauses[0].types[0]');
  this.text = node.getText().trim();
  this.name = node.name ? symbol.getName() : '';
  this.nameStart = node.name.getFullStart() - node.getFullStart();
  this.doc = ts.displayPartsToString(symbol.getDocumentationComment());
  /** 注释中的`@xx`属性 */
  this.config = {};
  /** 函数 */
  this.methods = [];
  /** 继承声明 */
  this.heritage = extendClz ? {
    name: extendClz.expression.getText(),
    start: extendClz.getFullStart() - node.getFullStart(),
    end: extendClz.getEnd() - node.getFullStart(),
    typeArguments: TypeArgument(extendClz),
  } : undefined;
  /** 如 decalre/export/default */
  this.modifiers = _.map(node.modifiers, (mod) => mod.getText());
  symbol.getJsDocTags().forEach((t) => {
    setOrPush(this.config, t.name, t.text);
  });
  this.decorators = Decorator(node.decorators);
  node.members.forEach((member) => {
    if (member.kind !== SyntaxKind.MethodDeclaration) {
      return;
    }
    this.methods.push(new ClzMethod(member));
  });
}

/** Enumeration */
function Enum(node) {
  this.text = node.getText().trim();
  this.name = node.name.getText().trim();
  /** 枚举列表{key: value} */
  this.attrs = {};
  node.members.forEach((member, index) => {
    this.attrs[member.name.getText()] = member.initializer ? getLiteralValue(member.initializer) : index;
  });
}

/**
 * Import 语句
 * @property namespace import * as / import xx from
 * @property names import {} from
 */
function Import(node) {
  this.text = node.getText().trim();
  const namedBindings = node.importClause.namedBindings;
  if (!namedBindings) {
    this.namespace = node.importClause.getText().trim();
    return;
  }
  switch (namedBindings.kind) {
    case SyntaxKind.NamespaceImport: {
      this.namespace = namedBindings.name.getText().trim();
      break;
    }
    case SyntaxKind.NamedImports: {
      this.names = {};
      namedBindings.elements.forEach((e) => {
        this.names[_.get(e.propertyName, 'text', e.name.text)] = e.name.getText().trim();
      });
      break;
    }
  }
}

/** 暂未处理的类型 */
function Unrecognize(node) {
  this.name = node.name && node.name.getText().trim();
  this.text = node.getText().trim();
}

/** 结束符 */
function EndOfFile(node) {
  this.text = node.getText().trim();
}

/** 默认导出 */
function Export(node) {
  this.text = node.getText().trim();
}

const mapping = {
  [SyntaxKind.InterfaceDeclaration]: Interface,
  [SyntaxKind.TypeAliasDeclaration]: TypeAlias,
  [SyntaxKind.TypeReference]: TypeReference,
  [SyntaxKind.ClassDeclaration]: Clazz,
  [SyntaxKind.FunctionDeclaration]: Func,
  [SyntaxKind.FunctionType]: Func,
  [SyntaxKind.VariableDeclaration]: Variable,
  [SyntaxKind.VariableStatement]: Variable,
  [SyntaxKind.ModuleDeclaration]: SubModule,
  [SyntaxKind.UnionType]: UnionType,
  [SyntaxKind.ArrayType]: UnionType,
  [SyntaxKind.LiteralType]: Literal,
  [SyntaxKind.EndOfFileToken]: EndOfFile,
  [SyntaxKind.ImportDeclaration]: Import,
  [SyntaxKind.EnumDeclaration]: Enum,
  [SyntaxKind.ExportAssignment]: Export,
  [SyntaxKind.IntersectionType]: IntersectionType,
  [SyntaxKind.Parameter]: TypeAlias,
};

/** 公共处理 */
function visitNode(node) {
  let kind = node.kind;
  if (kind === SyntaxKind.PropertySignature) {
    kind = node.type.kind;
  }
  if (mapping[kind]) {
    return new mapping[kind](node);
  }
  return new Unrecognize(node);
}

module.exports = {
  Literal,
  UnionType,
  TypeAlias,
  Interface,
  TypeArgument,
  TypeReference,
  Variable,
  Func,
  SubModule,
  visitNode,
  Clazz,
  Import,
  Export,
  EndOfFile,
  IntersectionType,
};
