const ts = require('typescript');
const _ = require('lodash');

const { SyntaxKind } = ts;

const types = {
  [SyntaxKind.StringKeyword]: 'String',
  [SyntaxKind.NumberKeyword]: 'Number',
  [SyntaxKind.BooleanKeyword]: 'Boolean',
  [SyntaxKind.AnyKeyword]: 'Any',
};

/** 获取基础类型名 */
function getKindStr(node) {
  let kind = (node.type && node.type.kind) || node.kind;
  if (kind === SyntaxKind.ArrayType) {
    kind = (node.type || node).elementType.kind;
    node = (node.type || node).elementType;
  }
  if (kind === SyntaxKind.TypeReference) {
    // 单值; 数组
    const typeName = node.typeName || node.type.typeName;
    if (typeName.kind === SyntaxKind.QualifiedName) {
      return typeName.getText();
    }
    return _.get(node, 'type.typeName.text') || _.get(node, 'typeName.text');
  }
  if (_.includes([SyntaxKind.TypeAliasDeclaration, SyntaxKind.Parameter], kind)) {
    return node.type ? node.type.getText() : 'Any';
  }
  if (kind === SyntaxKind.TypeLiteral) {
    return 'Object';
  }
  return types[kind] || _.get(SyntaxKind[kind].match(/(\w+)Keyword/), 1);
}

/**
 * 获取字面量值
 * @param node
 * @param [kind] 可选, 默认 node.kind
 */
function getLiteralValue(node, kind) {
  switch (kind || node.kind) {
    case SyntaxKind.TrueKeyword: case SyntaxKind.FalseKeyword: {
      return node.getText() === 'true';
    }
    case SyntaxKind.StringLiteral: {
      // 字符串去除引号
      return _.get(node.getText().match(/['"](.+)['"]/), 1);
    }
    case SyntaxKind.NumericLiteral: {
      return Number(node.getText());
    }
    default: {
      return getKindStr(node);
    }
  }
}

/**
 * type alias 和 interface 是嵌套类型, 其他是简单类型,
 * @returns true | undefined(undefined可以让字段隐藏掉,false会保留字段)
 */
function isSimpleType(node) {
  let kind = node.kind || node.type.kind;
  if (kind === SyntaxKind.ArrayType) {
    kind = (node.type || node).elementType.kind;
  }
  return !_.includes([SyntaxKind.TypeReference, SyntaxKind.InterfaceDeclaration], kind);
}

function isArray(node) {
  return _.includes([SyntaxKind.TupleType, SyntaxKind.ArrayType], (node.type && node.type.kind) || node.kind);
}

/** 是否是 export 的字面量 */
function isExport(node) {
  return ts.getCombinedModifierFlags(node) & ts.ModifierFlags.Export;
}

function getRealKind(node) {
  if (node.kind === SyntaxKind.ArrayType) {
    return node.elementType.kind;
  }
  if (node.kind === SyntaxKind.PropertySignature) {
    return node.type.kind === SyntaxKind.ArrayType ? node.type.elementType.kind: node.type.kind;
  }
  if (node.kind === SyntaxKind.TypeAliasDeclaration && node.type.kind === SyntaxKind.ArrayType) {
    return node.type.elementType.kind;
  }
  if (node.type && node.type.kind) {
    return node.type.kind
  }
  return node.kind;
}

/** 给obj的k设置值, 如果已存在则改为数组 */
function setOrPush(obj, k, v) {
  if (obj[k]) {
    obj[k] = [].concat(obj[k], v);
  } else {
    obj[k] = v;
  }
}

module.exports = {
  getKindStr,
  isSimpleType,
  setOrPush,
  isExport,
  isArray,
  getLiteralValue,
  getRealKind,
};
